/*
KeePass for BlackBerry

Copyright 2007 Fairview 5 Engineering, LLC <george.joseph@fairview5.com>

KeePass for J2ME
Copyright 2007 Naomaru Itoi <nao@phoneid.org>

Java clone of KeePass - A KeePass file viewer for Java
Copyright 2006 Bill Zwicky <billzwicky@users.sourceforge.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 2

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package com.fairview5.keepassbb.kpobjects;

// PhoneID
import java.io.IOException;
import java.util.Calendar;
import java.util.Date;

import net.rim.device.api.crypto.RandomSource;
import net.rim.device.api.io.Base64InputStream;
import net.rim.device.api.io.Base64OutputStream;
import net.rim.device.api.util.Arrays;

/**
 * Structure containing information about one entry. <p/>
 * 
 * <PRE>
 * One entry: [FIELDTYPE(FT)][FIELDSIZE(FS)][FIELDDATA(FD)]
 * [FT+FS+(FD)][FT+FS+(FD)][FT+FS+(FD)][FT+FS+(FD)][FT+FS+(FD)]...
 * &lt;p/&gt;
 * [ 2 bytes] FIELDTYPE
 * [ 4 bytes] FIELDSIZE, size of FIELDDATA in bytes
 * [ n bytes] FIELDDATA, n = FIELDSIZE
 * &lt;p/&gt;
 * Notes:
 * - Strings are stored in UTF-8 encoded form and are null-terminated.
 * - FIELDTYPE can be one of the FT_ constants.
 * </PRE>
 * 
 * @author Naomaru Itoi <nao@phoneid.org>
 * @author Bill Zwicky <wrzwicky@pobox.com>
 * @author Dominik Reichl <dominik.reichl@t-online.de>
 */
public final class PwEntry {

	public static final int BUF_SIZE = 124;

	public String uuid;
	public int parentGroupId;
	public int imageId;

	public String title;
	public String url;
	public String username;

	public String password;

	public String additional;

	public Date tCreation;
	public Date tLastMod;
	public Date tLastAccess;
	public Date tExpire;

	/**
	 * A string describing what is in pBinaryData
	 */
	public String binaryDesc;
	private byte[] binaryData;

	public PwEntry() {
	}

	public PwEntry(PwGroup pwg) {
		tCreation = new Date();
		tLastMod = new Date();
		tLastAccess = new Date();
		Calendar c = Calendar.getInstance();
		c.set(Calendar.YEAR, 2999);
		tExpire = c.getTime();
		parentGroupId = pwg.groupId;
		byte[] temp = new byte[16];
		RandomSource.getBytes(temp);
		try {
			uuid = Base64OutputStream.encodeAsString(temp, 0, temp.length, false, false);
		} catch (IOException e) {
		}
	}

	public void wipe() {
		parentGroupId = 0;
		imageId = 0;
		title = null;
		url = null;
		username = null;
		password = null;
		additional = null;
		tCreation = null;
		tLastMod = null;
		tLastAccess = null;
		tExpire = null;
		binaryDesc = null;
		binaryData = null;
		uuid = null;
	}

	public void delete(PwManager pwm) {
		pwm.getEntriesTable().remove(uuid);
		PwGroup pwg = pwm.getGroup(parentGroupId);
		pwg.childEntries.remove(uuid);
		wipe();
		pwm.setDirty(true);
	}

	public void add(PwManager pwm) {
		int err = 0;
		try {
			PwGroup pwg = pwm.getGroup(parentGroupId);
			err++;
			pwg.childEntries.put(uuid, this);
			err++;
			pwm.getEntriesTable().put(uuid, this);
			err++;
			pwm.setDirty(true);
		} catch (Exception e) {
			throw new RuntimeException("Unable to add entry: " + e.getMessage());
		}
	}

	public void copy(PwEntry toEntry) {
		toEntry.parentGroupId = parentGroupId;
		toEntry.imageId = imageId;
		toEntry.title = title;
		toEntry.url = url;
		toEntry.username = username;
		toEntry.password = password;
		toEntry.additional = additional;
		toEntry.tCreation = tCreation;
		toEntry.tLastMod = tLastMod;
		toEntry.tLastAccess = tLastAccess;
		toEntry.tExpire = tExpire;
		toEntry.binaryDesc = binaryDesc;
		toEntry.binaryData = binaryData;
	}

	/**
	 * @return the actual password byte array.
	 */
	public String getPassword() {
		return password;
	}

	/**
	 * Securely erase old password before copying new.
	 */
	public void setPassword(String pw) {
		if (password != null) {
			password = null;
		}
		password = pw;
	}

	/**
	 * @return the actual binaryData byte array.
	 */
	public byte[] getBinaryData() {
		return binaryData;
	}

	/**
	 * Securely erase old data before copying new.
	 */
	public void setBinaryData(byte[] buf, int offset, int len) {
		if (binaryData != null) {
			Arrays.zero(binaryData);
			binaryData = null;
		}
		binaryData = new byte[len];
		System.arraycopy(buf, offset, binaryData, 0, len);
	}

	public String toString() {
		return (title);
	}

	public void serialize(KdbOutputStream baos) {
		try {

			baos.writeByteArrayField(1, Base64InputStream.decode(uuid));
			baos.writeIntField(2, parentGroupId);
			baos.writeIntField(3, imageId);
			if (title != null)
				baos.writeStringField(4, title);
			if (url != null)
				baos.writeStringField(5, url);
			if (username != null)
				baos.writeStringField(6, username);
			if (password != null)
				baos.writeStringField(7, password);
			if (additional != null)
				baos.writeStringField(8, additional);
			baos.writeTimeField(9, tCreation);
			baos.writeTimeField(10, tLastMod);
			baos.writeTimeField(11, tLastAccess);
			baos.writeTimeField(12, tExpire);
			if (binaryDesc != null)
				baos.writeStringField(13, binaryDesc);
			if (binaryData != null)
				baos.writeByteArrayField(14, binaryData);
			baos.writeEOF();

		} catch (Exception e) {
			throw new RuntimeException("Unable to serialize entry: " + e.getMessage());
		}
	}

	public void unserialize(KdbInputStream bais) {
		try {
			int fieldType = bais.readShort();
			int fieldLength = bais.readInt();

			while (fieldType != 0xffff) {
				readEntryField(fieldType, fieldLength, bais);
				fieldType = bais.readShort();
				fieldLength = bais.readInt();
			}
		} catch (Exception e) {
			throw new RuntimeException("Unable to unserialize entry: " + e.getMessage());
		}
	}

	private void readEntryField(int fieldType, int fieldLength, KdbInputStream bais) throws IOException {

		switch (fieldType) {
		case 0x0000:
			// Ignore field
			break;
		case 0x0001:
			byte[] temp = new byte[16];
			bais.read(temp, 0, temp.length);
			uuid = Base64OutputStream.encodeAsString(temp, 0, temp.length, false, false);
			break;
		case 0x0002:
			parentGroupId = bais.readInt();
			break;
		case 0x0003:
			imageId = bais.readInt();
			break;
		case 0x0004:
			title = bais.readString(fieldLength);
			break;
		case 0x0005:
			url = bais.readString(fieldLength);
			break;
		case 0x0006:
			username = bais.readString(fieldLength);
			break;
		case 0x0007:
			password = bais.readString(fieldLength);
			break;
		case 0x0008:
			additional = bais.readString(fieldLength);
			break;
		case 0x0009:
			tCreation = bais.readTime();
			break;
		case 0x000A:
			tLastMod = bais.readTime();
			break;
		case 0x000B:
			tLastAccess = bais.readTime();
			break;
		case 0x000C:
			tExpire = bais.readTime();
			break;
		case 0x000D:
			binaryDesc = bais.readString(fieldLength);
			break;
		case 0x000E:
			if (binaryData != null)
				Arrays.zero(binaryData);
			binaryData = new byte[fieldLength];
			bais.read(binaryData, 0, binaryData.length);
			break;
		}
	}

}
