/*
KeePass for BlackBerry

Copyright 2007 Fairview 5 Engineering, LLC <george.joseph@fairview5.com>

KeePass for J2ME
Copyright 2007 Naomaru Itoi <nao@phoneid.org>

Java clone of KeePass - A KeePass file viewer for Java
Copyright 2006 Bill Zwicky <billzwicky@users.sourceforge.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 2

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package com.fairview5.keepassbb.kpobjects;

import java.io.IOException;
import java.util.Calendar;
import java.util.Date;

import net.rim.device.api.synchronization.UIDGenerator;

/**
 * @author Naomaru Itoi <nao@phoneid.org>
 * @author Bill Zwicky <wrzwicky@pobox.com>
 * @author Dominik Reichl <dominik.reichl@t-online.de>
 */
public final class PwGroup {

	public static final int BUF_SIZE = 124;

	// for tree traversing
	public SortedHashtable childGroups = new SortedHashtable();
	public SortedHashtable childEntries = new SortedHashtable();

	public int parentGroupId;
	public int groupId;
	public int imageId;
	public String name;

	public Date tCreation;
	public Date tLastMod;
	public Date tLastAccess;
	public Date tExpire;

	public int level; // short

	/**
	 * Used by KeePass internally, don't use
	 */
	public int flags;

	public PwGroup() {

	}

	public PwGroup(PwGroup parent, String name) {
		this(parent, name, 48);
	}

	public PwGroup(PwGroup parent, String name, int icon) {
		this.name = name;
		this.parentGroupId = parent.groupId;
		groupId = UIDGenerator.getUID();
		tCreation = new Date();
		tLastMod = new Date();
		tLastAccess = new Date();
		Calendar c = Calendar.getInstance();
		c.set(Calendar.YEAR, 2999);
		tExpire = c.getTime();
		level = parent.level + 1;
		imageId = icon;
	}

	public PwGroup(PwGroup parent, String name, int icon, int lev) {
		this.name = name;
		if (parent != null)
			this.parentGroupId = parent.groupId;
		else
			this.parentGroupId = -1;
		groupId = UIDGenerator.getUID();
		tCreation = new Date();
		tLastMod = new Date();
		tLastAccess = new Date();
		Calendar c = Calendar.getInstance();
		c.set(Calendar.YEAR, 2999);
		tExpire = c.getTime();
		level = lev;
		imageId = icon;
	}

	public void wipe() {
		groupId = -1;
		imageId = 0;
		name = null;
		tCreation = null;
		tLastMod = null;
		tLastAccess = null;
		tExpire = null;
		level = -1;
		flags = 0;
		if (childGroups != null) {
			PwGroup[] gps = childGroups.sortedGroupObjects();
			for (int i = 0; i < gps.length; i++)
				gps[i].wipe();
			childGroups.clear();
		}
		if (childEntries != null) {
			PwEntry[] es = childEntries.sortedEntryObjects();
			for (int i = 0; i < es.length; i++)
				es[i].wipe();
			childEntries.clear();
		}

	}

	public void add(PwManager pwm) {
		try {
			PwGroup pwg;
			if (level == 0)
				pwg = pwm.rootGroup;
			else
				pwg = pwm.getGroup(parentGroupId);
			pwg.childGroups.put(groupId, this);
			pwm.getGroupsTable().put(groupId, this);
			pwm.setDirty(true);
		} catch (Exception e) {
			throw new RuntimeException("Unable to add group: " + e.getMessage());
		}
	}

	public void delete(PwManager pwm) {
		try {
			if (childGroups != null) {
				PwGroup[] cg = childGroups.sortedGroupObjects();
				for (int i = 0; i < cg.length; i++) {
					cg[i].delete(pwm);
					cg[i].wipe();
				}
			}
			if (childEntries != null) {
				PwEntry[] ce = childEntries.sortedEntryObjects();
				for (int i = 0; i < ce.length; i++) {
					ce[i].delete(pwm);
					ce[i].wipe();
				}
			}

			PwGroup parent;
			if (level == 0)
				parent = pwm.rootGroup;
			else
				parent = pwm.getGroup(parentGroupId);
			parent.childGroups.remove(groupId);
			pwm.getGroupsTable().remove(groupId);
			wipe();
			pwm.setDirty(true);
		} catch (Exception e) {
			throw new RuntimeException("Unable to delete group: " + e.getMessage());
		}
	}

	public void copy(PwGroup toGroup) {
		toGroup.name = name;
		toGroup.parentGroupId = parentGroupId;
		toGroup.imageId = imageId;
		toGroup.level = level;
		toGroup.flags = flags;
		toGroup.tLastAccess = tLastAccess;
		toGroup.tCreation = tCreation;
		toGroup.tLastMod = tLastMod;
		toGroup.tExpire = tExpire;
	}

	public String toString() {
		return name;
	}

	public void serialize(KdbOutputStream baos) {
		try {
			baos.writeIntField(1, groupId);
			baos.writeStringField(2, name);
			baos.writeTimeField(3, tCreation);
			baos.writeTimeField(4, tLastMod);
			baos.writeTimeField(5, tLastAccess);
			baos.writeTimeField(6, tExpire);
			baos.writeIntField(7, imageId);
			baos.writeShortField(8, level);
			baos.writeIntField(9, flags);
			baos.writeEOF();
		} catch (Exception e) {
			throw new RuntimeException("Unable to serialize group: " + e.getMessage());
		}
	}

	public int deepSerialize(KdbOutputStream baos) {
		try {
			int rc = 1;
			serialize(baos);
			if (childGroups == null)
				return (1);
			PwGroup[] gps = childGroups.sortedGroupObjects();
			if (gps != null) {
				for (int i = 0; i < gps.length; i++) {
					rc += gps[i].deepSerialize(baos);
				}
			}
			return (rc);
		} catch (Exception e) {
			throw new RuntimeException("Unable to deep serialize group: " + e.getMessage());
		}
	}

	public void unserialize(KdbInputStream bais) {
		try {
			int fieldType = bais.readShort();
			int fieldLength = bais.readInt();

			while (fieldType != 0xffff) {
				readGroupField(fieldType, fieldLength, bais);
				fieldType = bais.readShort();
				fieldLength = bais.readInt();
			}
		} catch (Exception e) {
			throw new RuntimeException("Unable to unserialize group: " + e.getMessage());
		}
	}

	void readGroupField(int fieldType, int fieldLength, KdbInputStream bais) throws IOException {
		switch (fieldType) {
		case 0x0000:
			// Ignore field
			break;
		case 0x0001:
			groupId = bais.readInt();
			break;
		case 0x0002:
			name = bais.readString(fieldLength);
			break;
		case 0x0003:
			tCreation = bais.readTime();
			break;
		case 0x0004:
			tLastMod = bais.readTime();
			break;
		case 0x0005:
			tLastAccess = bais.readTime();
			break;
		case 0x0006:
			tExpire = bais.readTime();
			break;
		case 0x0007:
			imageId = bais.readInt();
			break;
		case 0x0008:
			level = bais.readShort();
			break;
		case 0x0009:
			flags = bais.readInt();
			break;
		}
	}

}
