/*
KeePass for BlackBerry

Copyright 2007 Fairview 5 Engineering, LLC <george.joseph@fairview5.com>

KeePass for J2ME
Copyright 2007 Naomaru Itoi <nao@phoneid.org>

Java clone of KeePass - A KeePass file viewer for Java
Copyright 2006 Bill Zwicky <billzwicky@users.sourceforge.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 2

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package com.fairview5.keepassbb.kpobjects;

// Java
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.Stack;

import net.rim.device.api.crypto.*;
import net.rim.device.api.io.NoCopyByteArrayOutputStream;
import net.rim.device.api.synchronization.SyncObject;
import net.rim.device.api.ui.component.Dialog;
import net.rim.device.api.util.Arrays;

import com.fairview5.keepassbb.Options;
import com.fairview5.keepassbb.Utils;
import com.fairview5.keepassbb.file.ByteArray;
import com.fairview5.keepassbb.file.KFile;
import com.fairview5.keepassbb.ui.PasswordDialog;
import com.fairview5.keepassbb.ui.ProgressDialog;
import com.fairview5.twofish.TwofishImpl;

/**
 * @author Naomaru Itoi <nao@phoneid.org>
 * @author Bill Zwicky <wrzwicky@pobox.com>
 * @author Dominik Reichl <dominik.reichl@t-online.de>
 */
public final class PwManager implements SyncObject {

	static {
		TwofishImpl.init();
	}

	// Descriptive name for database, used in GUI.
	public static String name = "KeePass database";
	public static String[] encryptionAlgorithms = { "AES", "Twofish" };
	public static PwManager pmCurrent;

	private boolean isDirty;
	private boolean isLegacy;

	PwEntry metaInfo;
	private SortedHashtable entriesTable = new SortedHashtable();
	private SortedHashtable groupsTable = new SortedHashtable();

	byte[] masterKey = new byte[32];
	byte[] masterSeed2 = new byte[32];
	byte[] transformedMasterKey = new byte[32];
	byte[] finalKey = new byte[32];
	int algorithm;
	int numKeyEncRounds;
	public PwGroup rootGroup;
	String timezone;

	private String filename;

	void createRootGroup() {
		rootGroup = new PwGroup();
		rootGroup.level = -1;
		rootGroup.groupId = -1;
		rootGroup.childGroups = new SortedHashtable();
		rootGroup.childEntries = new SortedHashtable();
	}

	public static boolean isValid() {
		return pmCurrent != null;
	}

	public PwManager() {
		createRootGroup();
	}

	PwManager(byte[] ba) {
		createRootGroup();
		unserializeWithPrompt(ba);
		if (pmCurrent == null)
			pmCurrent = this;
	}

	PwManager(byte[] ba, byte[] mk, String tz) {
		createRootGroup();
		timezone = tz;
		setMasterKey(mk);
		this.unserialize(ba, null);
		if (pmCurrent == null)
			pmCurrent = this;
	}

	PwManager(byte[] ba, String password) throws IOException {
		createRootGroup();
		timezone = null;
		String kfs = Options.getStringOption(Options.OPTION_DEFAULTKEYFILEURL);
		KFile kf = null;
		if (kfs != null) {
			kf = new KFile(kfs);
		}
		byte[] mk = createMasterKey(password, new KeyFile(kf.readContents()));
		
		setMasterKey(mk);
		
		this.unserialize(ba, null);
		
		if (pmCurrent == null)
			pmCurrent = this;
	}

	PwManager(byte[] ba, String password, KFile keyFile) throws IOException {
		createRootGroup();
		timezone = null;
		byte[] mk = createMasterKey(password, new KeyFile(keyFile.readContents()));
		
		setMasterKey(mk);
		
		this.unserialize(ba, null);
		
		if (pmCurrent == null)
			pmCurrent = this;
	}
	
	public PwManager(byte[] ba, byte[] mk) {
		this(ba, mk, (String) null);
	}

	public boolean load(KFile file) throws IOException {
		try {
			createRootGroup();
			if (!this.unserializeWithPrompt(file.readContents())) {
				return false;
			}
			setFilename(file.getFullName());
			setDirty(false);
			setLegacy(false);
		} finally {
			Utils.closeStream(file);
		}
		return true;
	}

	public void setMasterKey(String pw) {
		SHA256Digest md = new SHA256Digest();
		md.update(pw.getBytes());
		md.getDigest(masterKey, 0);
	}

	public void setMasterKey(byte[] mk) {
		System.arraycopy(mk, 0, masterKey, 0, masterKey.length);
	}

	public void reset() {
		try {
			PwGroup[] gps = getGroupsTable().sortedGroupObjects();
			if (gps != null)
				for (int i = 0; i < gps.length; i++)
					gps[i].wipe();

			PwEntry[] es = getEntriesTable().sortedEntryObjects();
			if (es != null)
				for (int i = 0; i < es.length; i++)
					es[i].wipe();

			getGroupsTable().clear();
			getEntriesTable().clear();
			metaInfo = null;
			Arrays.zero(masterKey);
			Arrays.zero(masterSeed2);
			Arrays.zero(transformedMasterKey);
			Arrays.zero(finalKey);
			algorithm = 0;
			numKeyEncRounds = 0;
			rootGroup = null;
		} catch (Throwable e) {
			throw new RuntimeException("Unable to reset PwManager: " + e.getMessage());
		}

	}

	public String toString() {
		return name;
	}
	
	public synchronized void save() throws IOException {
		save(null);
	}

	public void save(boolean progressGauge) throws IOException {
		ProgressDialog pd = null;
		if (progressGauge) {
			pd = new ProgressDialog("");
			pd.show();
		}
		save(pd);
		if (pd != null) pd.close();
	}
	

	public synchronized void save(ProgressDialog gauge) throws IOException {
		KFile fc = null;
		try {
			ByteArray ba = serialize(gauge);
			fc = new KFile(getFilename());
			if (fc.exists())
				fc.truncate();
			else
				fc.create();
			fc.write(ba.byteArray, 0, ba.length);
			fc.flush();
			setDirty(false);
		} finally {
			Utils.closeStream(fc);
		}
	}

	public void saveAs(String filename, ProgressDialog gauge) throws IOException {
		saveAs(new KFile(filename), gauge);
	}

	public void saveAs(KFile f, boolean progressGauge) throws IOException {
		ProgressDialog pd = null;
		if (progressGauge) {
			pd = new ProgressDialog("");
			pd.show();
		}
		saveAs(f, pd);
		if (pd != null) pd.close();
	}
	
	public synchronized void saveAs(KFile f, ProgressDialog gauge) throws IOException {
		try {
			ByteArray ba = serialize(gauge);
			if (f.exists())
				f.truncate();
			else
				f.create();
			f.write(ba.byteArray, 0, ba.length);
			f.flush();
			setFilename(f.getFullName());
			setDirty(false);
		} finally {
			Utils.closeStream(f);
		}
	}

	public PwGroup getGroup(int groupId) {
		return (getGroupsTable().get(groupId));
	}

	public PwEntry getEntry(String uuid) {
		return (getEntriesTable().get(uuid));
	}

	public boolean unserializeWithPrompt(final byte[] filebuf) {
		try {
			PasswordDialog pw = new PasswordDialog();
			int rc = pw.doModal();
			if (rc != Dialog.OK)
				return (false);
			String password = pw.getPassword();
			KeyFile kf = pw.getKeyFile();

			if ((password == null || password.length() <= 0) && (kf == null))
				return (false);

			SHA256Digest mdPW = null;
			if (password.length() > 0) {
				mdPW = new SHA256Digest();
				mdPW.update(password.getBytes());
			}

			SHA256Digest mdKF = null;

			if (kf != null) {
				if (mdPW != null) {
					mdKF = new SHA256Digest();
					mdKF.update(mdPW.getDigest());
					mdKF.update(kf.getKey());
					mdKF.getDigest(masterKey, 0);
				} else {
					System.arraycopy(kf.getKey(), 0, masterKey, 0, 32);
				}
			} else {
				mdPW.getDigest(masterKey, 0);
			}

			final ProgressDialog pg = new ProgressDialog("Preparing master key...");

			final Throwable[] ex = new Throwable[1];

			Thread t = new Thread(new Runnable() {
				public void run() {
					try {
						unserialize(filebuf, pg);
					} catch (Throwable e) {
						ex[0] = e;
					}
					pg.close();
				}
			});
			t.start();
			pg.doModal();
			if (ex[0] != null)
				throw ex[0];
			return (true);
		} catch (Throwable e) {
			throw new RuntimeException(e.getMessage());
		}
	}

	public void unserialize(byte[] filebuf, ProgressDialog gauge) {

		if (filebuf == null || filebuf.length < PwDbHeader.BUF_SIZE)
			throw new RuntimeException("File too short for header");

		SHA256Digest md;
		KdbInputStream bais = new KdbInputStream(filebuf, timezone);
		KdbInputStream baispt = null;
		
		PwDbHeader hdr = new PwDbHeader();

		try {
			hdr.unserialize(bais);

			if ((hdr.signature1 != PwDbHeader.PWM_DBSIG_1) || (hdr.signature2 != PwDbHeader.PWM_DBSIG_2)) {
				throw new RuntimeException("Bad database file signature");
			}
			if ((hdr.version & 0xffffff00) != (PwDbHeader.PWM_DBVER_DW & 0xffffff00)) {
				throw new RuntimeException("Bad database file version");
			}
			if ((hdr.flags & PwDbHeader.PWM_FLAG_RIJNDAEL) != 0) {
				algorithm = PwDbHeader.ALGO_AES;
			} else if ((hdr.flags & PwDbHeader.PWM_FLAG_TWOFISH) != 0) {
				algorithm = PwDbHeader.ALGO_TWOFISH;
			} else {
				throw new RuntimeException("Unknown algorithm.");
			}
		} catch (Throwable t) {
			Utils.closeStream(bais);
			throw new RuntimeException("header: " + t.getMessage());
		}

		DecryptorInputStream dis = null;
		NoCopyByteArrayOutputStream tempout = null;
		ByteArrayInputStream tempin = null;
		try {
			numKeyEncRounds = hdr.numKeyEncRounds;
			name = "KeePass Password Manager";

			System.arraycopy(hdr.masterSeed2, 0, masterSeed2, 0, masterSeed2.length);
			transformMasterKey(gauge);
			md = new SHA256Digest();
			md.update(hdr.masterSeed, 0, hdr.masterSeed.length);
			md.update(transformedMasterKey, 0, transformedMasterKey.length);
			md.getDigest(finalKey, 0, true);
			
			InitializationVector iv = new InitializationVector(hdr.encryptionIV);
			
			SymmetricKey sk = SymmetricKeyFactory.getInstance(encryptionAlgorithms[algorithm] + "_256", finalKey, 0,
					finalKey.length);
			dis = DecryptorFactory.getDecryptorInputStream(sk, bais, encryptionAlgorithms[algorithm] + "/CBC/PKCS5", iv);

			tempout = new NoCopyByteArrayOutputStream(filebuf.length);
			byte[] xfer = new byte[1024];
			int br = 0;
			int tbr = 0;
			
			if (gauge != null) {
				gauge.setValue(0);
				gauge.setText("Decrypting...");
			}
			while ((br = dis.read(xfer, 0, 1024)) > 0) {
				tempout.write(xfer, 0, br);
				tbr += br;
				int pct = ((tbr * 100) / filebuf.length);
				if (gauge != null && pct % 5 == 0) {
					gauge.setValue(pct);
				}
			}
			byte[] clearData = tempout.getByteArray();
			SHA256Digest testhash = new SHA256Digest();
			testhash.update(clearData, 0, tempout.size());
			if (Arrays.equals(testhash.getDigest(), hdr.contentsHash) == false) {
				throw new RuntimeException("Hash mismatch");
			}
			tempin = new ByteArrayInputStream(clearData, 0, tempout.size());
			baispt = new KdbInputStream(tempin, timezone);
		} catch (Throwable e) {
			Utils.logger(e.toString()+":"+e.getMessage());
			Utils.closeStream(tempin);
			Utils.closeStream(baispt);
			throw new RuntimeException("Bad password or corrupted file.");
		} finally {
			Utils.closeStream(tempout);
			Utils.closeStream(dis);
			Utils.closeStream(bais);
		}

		try {
			if (gauge != null) {
				gauge.setValue(0);
				gauge.setText("Creating objects...");
			}
			
			rootGroup.wipe();
			Stack gs = new Stack();
			gs.push(rootGroup);
			PwGroup currentParent = rootGroup;
			PwGroup gp;
			for (int i = 0; i < hdr.numGroups; i++) {
				gp = new PwGroup();
				gp.unserialize(baispt);
				while (gp.level <= currentParent.level) {
					gs.pop();
					currentParent = (PwGroup) gs.peek();
				}
				gp.parentGroupId = currentParent.groupId;
				gp.add(this);
				gs.push(gp);
				currentParent = gp;
				
				int pct = ((i * 100) / (hdr.numGroups + hdr.numEntries));
				if (gauge != null && pct % 5 == 0) {
					gauge.setValue(pct);
				}
				
			}
		} catch (Throwable e) {
			throw new RuntimeException("groups: " + e.toString());
		}

		try {
			PwEntry ent;
			for (int i = 0; i < hdr.numEntries; i++) {
				ent = new PwEntry();
				ent.unserialize(baispt);
				if (ent.title != null && ent.title.equals("Meta-Info")) {
					metaInfo = ent;
				} else
					ent.add(this);
				int pct = (((i + hdr.numGroups) * 100) / (hdr.numGroups + hdr.numEntries));
				if (gauge != null && pct % 5 == 0) {
					gauge.setValue(pct);
				}
			}
		} catch (Throwable e) {
			throw new RuntimeException("entries: " + e.toString());
		}
		Utils.saveToRuntimeStore(masterKey, PwSyncManager.GUID_MASTERKEY);
	}

	public ByteArray serialize(ProgressDialog gauge) {

		KdbOutputStream baos = null;

		baos = new KdbOutputStream(timezone, 2048);
		try {
			baos.write(new byte[PwDbHeader.BUF_SIZE]);
		} catch (IOException e2) {
			e2.printStackTrace();
		}
		int cvg = getGroupsTable().size();
		int cve = getEntriesTable().size();
		int rc = 0;
		
		if (gauge != null) {
			gauge.setValue(0);
			gauge.setText("Serializing objects...");
		}

		try {
			if (rootGroup != null && rootGroup.childGroups != null) {
				PwGroup[] pwg = rootGroup.childGroups.sortedGroupObjects();
				if (pwg != null) {
					for (int i = 0; i < pwg.length; i++) {
						rc += pwg[i].deepSerialize(baos);
						int pct = ((i * 100) / (cvg+cve));
						if (gauge != null && pct % 5 == 0) {
							gauge.setValue(pct);
						}
					}
				}
				if (cvg != rc)
					throw new RuntimeException("Group serialization mismatch: " + cvg + ":" + rc);
			}
		} catch (Throwable e) {
			Utils.closeStream(baos);
			throw new RuntimeException("groups " + e.toString());
		}


		try {
			if (getEntriesTable() != null) {
				PwEntry[] pwe = getEntriesTable().sortedEntryObjects();
				if (pwe != null) {
					for (int i = 0; i < pwe.length; i++) {
						pwe[i].serialize(baos);
						int pct = (((i+cvg) * 100) / (cvg+cve));
						if (gauge != null && pct % 5 == 0) {
							gauge.setValue(pct);
						}
					}
				}
			}
		} catch (Throwable e) {
			Utils.closeStream(baos);
			throw new RuntimeException("entries" + e.toString());
		}

		try {
			if (metaInfo != null) {
				metaInfo.serialize(baos);
				cve++;
			}
		} catch (Throwable e) {
			Utils.closeStream(baos);
			throw new RuntimeException("meta " + e.toString());
		}

		int pcSize = baos.size()-PwDbHeader.BUF_SIZE;
		try {
			baos.write(new byte[32]);
		} catch (IOException e1) {}
		
		byte[] plainContent = baos.getByteArray();
		Utils.closeStream(baos);

		SHA256Digest md = new SHA256Digest();
		PwDbHeader ph = new PwDbHeader();
		try {
			ph.signature1 = PwDbHeader.PWM_DBSIG_1;
			ph.signature2 = PwDbHeader.PWM_DBSIG_2;
			ph.flags = PwDbHeader.PWM_FLAG_SHA2;
			if (algorithm == PwDbHeader.ALGO_AES)
				ph.flags |= PwDbHeader.PWM_FLAG_RIJNDAEL;
			else if (algorithm == PwDbHeader.ALGO_TWOFISH)
				ph.flags |= PwDbHeader.PWM_FLAG_TWOFISH;
			ph.version = PwDbHeader.PWM_DBVER_DW;
			ph.numGroups = cvg;
			ph.numEntries = cve;
			ph.numKeyEncRounds = numKeyEncRounds;

			RandomSource.getBytes(ph.masterSeed, 0, ph.masterSeed.length);
			RandomSource.getBytes(ph.encryptionIV, 0, ph.encryptionIV.length);
			System.arraycopy(masterSeed2, 0, ph.masterSeed2, 0, ph.masterSeed2.length);
			
			md.reset();
			md.update(plainContent, PwDbHeader.BUF_SIZE, pcSize);
			md.getDigest(ph.contentsHash, 0, true);
			
			KdbOutputStream kph = new KdbOutputStream(PwDbHeader.BUF_SIZE);
			ph.serialize(kph);
			System.arraycopy(kph.getByteArray(), 0, plainContent, 0, PwDbHeader.BUF_SIZE);
			kph.close();
			
			md.update(ph.masterSeed);
			md.update(transformedMasterKey);
			md.getDigest(finalKey, 0, true);
			Utils.saveToRuntimeStore(finalKey, PwSyncManager.GUID_FINALKEY);
			
		} catch (Throwable e) {
			throw new RuntimeException("header" + e.toString());
		}
		
		try {

			SymmetricKey sk = SymmetricKeyFactory.getInstance(encryptionAlgorithms[algorithm] + "_256", finalKey, 0,
					finalKey.length);
			
			BlockEncryptorEngine bes = EncryptorFactory.getBlockEncryptorEngine(sk, encryptionAlgorithms[algorithm] + "/CBC",
					new InitializationVector(ph.encryptionIV));
			
			PKCS5FormatterEngine p5 = new PKCS5FormatterEngine(bes);
			
			int st = PwDbHeader.BUF_SIZE;
			int bl = pcSize;
			int wr = 0;
			
			if (gauge != null) {
				gauge.setValue(0);
				gauge.setText("Encrypting...");
			}
			
			while (bl > 0) {
				int ibl = Math.min(p5.getInputBlockLength(), bl);
				boolean lastBlock = (bl <= ibl);
				int be = p5.formatAndEncrypt(plainContent, st, ibl, plainContent, st, lastBlock);
				st += ibl;
				bl -= ibl;
				wr+=be;
				
				int pct = ((wr * 100) / pcSize);
				if (gauge != null && pct % 5 == 0) {
					gauge.setValue(pct);
				}
				
			}
			pcSize = wr;
		} catch (Throwable e) {
			throw new RuntimeException("crypto " + e.toString());
		}
		return new ByteArray(plainContent, pcSize+PwDbHeader.BUF_SIZE);
	}

	public void transformMasterKey(ProgressDialog gauge) {

		try {
			int i;

			AESKey ak = new AESKey(masterSeed2, 0, 256);
			AESEncryptorEngine aes = new AESEncryptorEngine(ak, 16);

			byte[][] temp = new byte[2][32];
			byte[] in = masterKey;
			byte[] out = temp[1];

			int incr = numKeyEncRounds / 20;
			for (i = 0; i < numKeyEncRounds; i++) {
				aes.encrypt(in, 0, out, 0);
				aes.encrypt(in, 16, out, 16);
				in = temp[(i + 1) % 2];
				out = temp[i % 2];

				if (gauge != null && (i % (incr)) == 0)
					gauge.setValue(i, numKeyEncRounds - 1);
			}
			// Hash once with SHA-256

			SHA256Digest md = new SHA256Digest();
			md.update(in);
			md.getDigest(transformedMasterKey, 0);
		} catch (Throwable e) {
			throw new RuntimeException(e.toString());
		}

	}
	
	public static byte[] createMasterKey(String password, KeyFile kf) {
		byte[] mk = new byte[32];
		
		SHA256Digest mdPW = null;
		if (password.length() > 0) {
			mdPW = new SHA256Digest();
			mdPW.update(password.getBytes());
		}

		SHA256Digest mdKF = null;
		if (kf != null) {
			if (mdPW != null) {
				mdKF = new SHA256Digest();
				mdKF.update(mdPW.getDigest());
				mdKF.update(kf.getKey());
				mdKF.getDigest(mk, 0);
			} else {
				System.arraycopy(kf, 0, mk, 0, 32);
			}
		} else {
			mdPW.getDigest(mk, 0);
		}
		return mk;
		
	}

	public boolean setMasterPassword() {

		DatabasePropertiesDialog dlg = new DatabasePropertiesDialog(numKeyEncRounds, algorithm);
		int rc = dlg.doModal();
		if (rc != Dialog.OK)
			return (false);
		String password = dlg.getPassword();
		KeyFile kf = dlg.getKeyFile();

		SHA256Digest mdPW = null;
		if (password.length() > 0) {
			mdPW = new SHA256Digest();
			mdPW.update(password.getBytes());
		}

		SHA256Digest mdKF = null;
		if (kf != null) {
			if (mdPW != null) {
				mdKF = new SHA256Digest();
				mdKF.update(mdPW.getDigest());
				mdKF.update(kf.getKey());
				mdKF.getDigest(masterKey, 0);
			} else {
				System.arraycopy(kf, 0, masterKey, 0, 32);
			}
		} else {
			mdPW.getDigest(masterKey, 0);
		}

		RandomSource.getBytes(masterSeed2, 0, masterSeed2.length);

		Utils.saveToRuntimeStore(masterKey, PwSyncManager.GUID_MASTERKEY);
		numKeyEncRounds = dlg.getKeyRounds();
		algorithm = dlg.getAlgorithm();

		final ProgressDialog pg = new ProgressDialog("Preparing master key...");
		final Exception[] ex = new Exception[1];
		Thread t = new Thread(new Runnable() {
			public void run() {
				try {
					transformMasterKey(pg);
				} catch (Exception e) {
					ex[0] = e;
				}
				pg.close();
			}
		});
		t.start();
		pg.show();
		if (ex[0] != null) {
			String msg = ex[0].getMessage();
			if (msg == null || msg.length() == 0)
				Dialog.alert(ex[0].toString());
			else
				Dialog.alert(msg);
			return (false);
		}

		return (true);
	}

	public void setGroupsTable(SortedHashtable groupsTable) {
		this.groupsTable = groupsTable;
	}

	public SortedHashtable getGroupsTable() {
		return groupsTable;
	}

	public void setFilename(String filename) {
		this.filename = filename;
	}

	public String getFilename() {
		return filename;
	}

	public void setEntriesTable(SortedHashtable entriesTable) {
		this.entriesTable = entriesTable;
	}

	public SortedHashtable getEntriesTable() {
		return entriesTable;
	}

	public static void setDefaultDirty(boolean isDirty) {
		pmCurrent.setDirty(isDirty);
	}

	public static boolean isDefaultDirty() {
		return isValid() && pmCurrent.isDirty();
	}

	public static void setDefaultLegacy(boolean isLegacy) {
		pmCurrent.setLegacy(isLegacy);
	}

	public static boolean isDefaultLegacy() {
		return isValid() && pmCurrent.isLegacy();
	}

	public void setDirty(boolean isDirty) {
		this.isDirty = isDirty;
	}

	public boolean isDirty() {
		return isDirty;
	}

	public void setLegacy(boolean isLegacy) {
		this.isLegacy = isLegacy;
	}

	public boolean isLegacy() {
		return isLegacy;
	}

	public void setUID(int uid) {
		this.uid = uid;
	}

	public int getUID() {
		return uid;
	}

	private int uid;
	public static final byte FIELD_TYPE_DATA = 0x01;
	public static final byte FIELD_TYPE_TZ = 0x00;

}
